'''
Module containing the function to run the glider ingest process
'''
# AUTHORS:
# Alec Krueger, Texas A&M University, Geochemical and Environmental Research Group, alecmkrueger@tamu.edu
# Sakib Mahmud, Texas A&M University, Geochemical and Environmental Research Group, sakib@tamu.edu
# Xiao Ge, Texas A&M University, Geochemical and Environmental Research Group, gexiao@tamu.edu

import xarray as xr
from pathlib import Path
from processor import Processor

def process(raw_data_source:Path|str,working_directory:Path|str,glider_number:str,mission_title:str,extensions:list,output_nc_filename:str,return_ds:bool=False) -> None|xr.Dataset:
    '''

    Function inputs:
    * raw_data_source (Path|str): Raw data source, from the glider SD card
    * working_directory (Path|str): Where you want the raw copy and processed data to be
    * glider_number (str): The number of the glider, for NetCDF metadata
    * mission_title (str): The mission title, for NetCDF metadata
    * extensions (list): The extensions you wish to process
    * output_nc_filename (str): The name of the output NetCDF file
    * return_ds (bool): If you would like the output dataset to be returned. Default = False

    Example Parameter inputs:
    * glider_number:str = '540'
    * mission_title:str = 'Mission_44'
    * extensions = ["DBD", "EBD"] or ["SBD", "TBD"]
    * raw_data_source = Path('test_data').resolve()
    * working_directory = Path('data').resolve()
    * output_nc_filename = 2024_mission_44.nc
    * return_ds = True
    '''
    if isinstance(raw_data_source,str):
        raw_data_source = Path(raw_data_source)
    if isinstance(working_directory,str):
        working_directory = Path(working_directory)

    if not raw_data_source.exists():
        raise ValueError(f'Raw data source directory does not exist: {raw_data_source}')

    if not working_directory.exists():
        raise ValueError(f'Working directory does not exist: {working_directory}')
    
    processor = Processor(raw_data_source=raw_data_source,working_directory=working_directory,glider_number=glider_number,
                          mission_title=mission_title,output_nc_filename=output_nc_filename,extensions=extensions)
    processor.process()

    if return_ds:
        return processor.ds_mission

# Example:
ds = process('540','Mission_44',extensions=['DBD','EBD'],raw_data_source=Path('../../test_data').resolve(),
             working_directory=Path('../../data').resolve(),output_nc_filename='test.nc',return_ds=True)






'''
Module containing the Processor class
'''
from attrs import define,field
import xarray as xr
from pathlib import Path
import os
import multiprocessing
from attrs import define,field
from concurrent.futures import ThreadPoolExecutor,as_completed
import platform

from utils import print_time,copy_file,rename_file,create_tasks,convert_file
from utils import process_sci_data,process_flight_data,add_gridded_data,add_global_attrs,length_validator

@define
class Processor:
    '''Class to process and contain information about the raw glider data ingest'''
    raw_data_source:Path
    working_directory:Path
    glider_number:str
    mission_title:str
    output_nc_filename:str
    extensions:list = field(default=['DBD','EBD'],validator=length_validator)
    data_sources:list = field(default=['Flight', 'Science'],validator=length_validator)
    max_workers:int|None = field(default=None)
    debug:bool = field(default=False)

    ds_mission:xr.Dataset = field(init=False)
    output_nc_path:Path = field(init=False)


    def __attrs_post_init__(self):
        self.output_nc_path = self.working_directory.joinpath('processed','nc',self.output_nc_filename)
        if self.max_workers is None:
            self.max_workers = multiprocessing.cpu_count()

    def print_time_debug(self,message):
        # Print the message if self.debug is true
        if self.debug:
            print_time(message)

    def create_directory(self):
        '''
        Create the directory that will contain the raw and processed data inside the user defined working directory
        Sets up the structure that the rest of the module uses to put the data in the correct spots
        '''
        # Create cache dir
        cache_path = Path('cache')
        cache_path.mkdir(exist_ok=True)
        # Define the two data type folders
        data_types = ['processed','raw_copy']
        # Define the three processed folders
        processed_data_types = ['Flight','nc','Science']
        # Define the raw data type folders by the file extension of the files to be stored
        raw_flight_extensions = ["DBD", "MBD", "SBD", "MLG"]
        raw_science_extensions = ["EBD", "NLG", "TBD", "NBD"]
        # Loop through the two data type folders
        for dtype in data_types:
            if dtype == 'processed':
                for processed_dtype in processed_data_types:
                    # Example directory being created: self.working_directory/processed/Flight
                    os.makedirs(self.working_directory.joinpath(dtype, processed_dtype), exist_ok=True)
            elif dtype == 'raw_copy':
                # Package Flight and Science with their respective data type folders (extensions)
                for data_source,extensions in zip(['Flight','Science'],[raw_flight_extensions,raw_science_extensions]):
                    for extension in extensions:
                        # Example directory being created: self.working_directory/raw_copy/Flight/DBD
                        os.makedirs(self.working_directory.joinpath(dtype ,data_source, extension), exist_ok=True)

    def delete_files_in_directory(self):
        '''
        Clean up the directory incase there are files that already exist.
        If you do not delete the files in the directory before running the rest of the script,
        then there will likely be data duplication and errors.
        So to be save, always clean up the directory using this function
        '''
        # Check if the user wishes to delete all files in the directory
        confirmation = input(f"Are you sure you want to delete all files in '{self.working_directory}' and its subdirectories? Type 'yes' to confirm, 'no' to continue without deleting files, press escape to cancel and end ")
        # If so then begin finding files
        if confirmation.lower() == 'yes':
            # clear cache
            cache_path = Path('cache').resolve()
            [os.remove(file) for file in cache_path.rglob('*.cac')]
            # clear all files in the given directory
            for root, _, files in os.walk(self.working_directory):
                file_count = len(files)
                for file in files:
                    file_path = os.path.join(root, file)
                    os.remove(file_path)
                if file_count > 0:
                    self.print_time_debug(f"Cleaned {root}, deleted {file_count} file(s).")
            self.print_time_debug("All files have been deleted")
        elif confirmation.lower() == 'no':
            self.print_time_debug('Continuing without deleting files, this may cause unexpected behaviours including data duplication')
        else:
            raise ValueError("Cancelling: If you did not press escape, ensure you type 'yes' or 'no'. ")   
         
    def copy_raw_data(self):
        '''
        Copy data from the memory card to the working directory using multithreading.
        We only work on the copied data and never the source data to keep the data safe
        '''
        self.print_time_debug('Copying Raw files')
        
        raw_output_data_dir = self.working_directory.joinpath('raw_copy')

        all_extensions = [["DBD", "MBD", "SBD", "MLG", "CAC"], ["EBD", "NLG", "TBD", "NBD", "CAC"]]
        
        tasks = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            for data_source, extensions in zip(self.data_sources, all_extensions):
                input_data_path = self.raw_data_source.joinpath(f'{data_source}_card')
                for file_extension in extensions:
                    # If the extension is CAC then we want to copy the files to the cache folder
                    if file_extension == 'CAC':
                        # Add cache to internal cache folder
                        output_data_path = Path('cache')
                    else:
                        output_data_path = raw_output_data_dir.joinpath(data_source, file_extension)
                    # Find all of the files with the file extension 
                    input_files = input_data_path.rglob(f'*.{file_extension}')
                    # Loop through the files with matching extensions
                    for input_file_path in input_files:
                        # Define where the file will be placed
                        output_file_path = output_data_path.joinpath(input_file_path.name)
                        # Append the input and output file paths to a list
                        tasks.append((input_file_path, output_file_path))
            # Queue the copy_file function using multiprocessing on the input and output file paths
            futures = [executor.submit(copy_file, input_file_path, output_file_path) for input_file_path, output_file_path in tasks]
            # Perform the multiprocessing
            for future in as_completed(futures):
                try:
                    future.result()  # Raise any exceptions that occurred
                except Exception as e:
                    self.print_time_debug(f"Error copying file: {e}")
        
        self.print_time_debug('Done Copying Raw files')
    def rename_binary_files(self):
        '''
        Rename files with extensions of DBD or EBD to contain date and glider name in the input data directory
        using multithreading.
        '''
        self.print_time_debug('Renaming dbd files')
        
        raw_copy_directory = self.working_directory.joinpath('raw_copy')
        current_os = platform.system()
        if current_os == 'Linux' or current_os == 'Darwin':
            rename_path = Path('rename_files.exe').resolve()
        elif current_os == 'Windows':
            rename_path = Path('windows_rename_files.exe').resolve()
        else:
            required_os_list = ['Windows','Linux','Darwin']
            raise ValueError(f"Unknown Operating System, got {current_os}, must be one of {required_os_list}")
        rename_files_path = rename_path

        tasks = []
        for extension in self.extensions:
            if extension is None:
                continue
            data_files = raw_copy_directory.rglob(f'*.{extension}')
            for file in data_files:
                tasks.append(file)

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(rename_file, rename_files_path, file) for file in tasks]
            
            for future in as_completed(futures):
                try:
                    future.result()  # Raise any exceptions that occurred
                except Exception as e:
                    self.print_time_debug(f"Error renaming file: {e}")

        self.print_time_debug("Done renaming dbd files")

    def convert_binary_to_ascii(self):
        '''Converts binary files to ascii in the input directory and saves them to the output directory'''
        self.print_time_debug('Converting to ascii')
        output_data_dir = self.working_directory.joinpath('processed')
        working_directory = self.working_directory.joinpath('raw_copy')
        
        # Define the Path object for where the binary2asc executable is
        current_os = platform.system()
        if current_os == 'Linux' or current_os == 'Darwin':
            binary2asc_path = Path('binary2asc.exe').resolve()
        elif current_os == 'Windows':
            binary2asc_path = Path('windows_binary2asc.exe').resolve()
        else:
            required_os_list = ['Windows','Linux','Darwin']
            raise ValueError(f"Unknown Operating System, got {current_os}, must be one of {required_os_list}")
        
        # Collect all files to be processed
        tasks = []
        for data_source, extension in zip(self.data_sources, self.extensions):
            tasks = create_tasks(working_directory,data_source,extension,output_data_dir,tasks,binary2asc_path)
        
        # Process files in parallel using ThreadPoolExecutor
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(convert_file, binary2asc_path, raw_file, ascii_file) for binary2asc_path, raw_file, ascii_file in tasks]
            for future in as_completed(futures):
                future.result()  # Raise any exceptions that occurred
        
        self.print_time_debug('Done Converting to ascii')

    def convert_ascii_to_dataset(self):
        '''Convert ascii data files into a single NetCDF file'''
        self.print_time_debug('Converting ascii to dataset')
        processed_directory = self.working_directory.joinpath('processed')

        science_data_dir:Path = processed_directory.joinpath('Science')
        flight_data_dir:Path = processed_directory.joinpath('Flight')

        glider_id = {'199':'Dora','307':'Reveille','308':'Howdy','540':'Stommel','541':'Sverdrup'}
        wmo_id = {'199':'unknown','307':'4801938','308':'4801915','540':'4801916','541':'4801924'}
        
        # Process Science Data
        ds_sci:xr.Dataset = process_sci_data(science_data_dir,glider_id,self.glider_number,wmo_id)

        # Make a copy of the science dataset
        ds_mission:xr.Dataset = ds_sci.copy()

        # Process Flight Data
        ds_fli:xr.Dataset = process_flight_data(flight_data_dir)

        # Add flight data to mission dataset
        ds_mission.update(ds_fli)

        # Add gridded data to mission dataset
        ds_mission = add_gridded_data(ds_mission)

        # Add attributes to the mission dataset
        self.ds_mission = add_global_attrs(ds_mission,mission_title=self.mission_title,wmo_id=wmo_id,glider=self.glider_number)

        self.print_time_debug('Finished converting ascii to dataset')

    def save_ds(self):
        '''Save the mission dataset to NetCDF'''
        self.print_time_debug(f'Saving dataset to {self.output_nc_path}')
        self.ds_mission.to_netcdf(self.output_nc_path)
        self.print_time_debug('Finished saving dataset')

    def process(self):
        '''Perform the processing of the raw glider data into a NetCDF file'''
        self.create_directory()
        self.delete_files_in_directory()
        self.copy_raw_data()
        self.rename_binary_files()
        self.convert_binary_to_ascii()
        self.convert_ascii_to_dataset()
        self.save_ds()

    

    



'''
Module containing utilities for the package
'''
# Import Packages
import numpy as np
import pandas as pd
import xarray as xr
from pathlib import Path
import datetime
import shutil
import subprocess
import os
import gsw
import multiprocessing
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from gridder import Gridder

# AUTHORS:
# Sakib Mahmud, Texas A&M University, Geochemical and Environmental Research Group, sakib@tamu.edu
# Xiao Ge, Texas A&M University, Geochemical and Environmental Research Group, gexiao@tamu.edu
# Alec Krueger, Texas A&M University, Geochemical and Environmental Research Group, alecmkrueger@tamu.edu


# Define functions and classes
def print_time(message):
    '''Add the current time to the end of a message'''
    # Get current time
    current_time = datetime.datetime.today().strftime('%H:%M:%S')
    # Add time to message
    whole_message = f'{message}: {current_time}'
    # Print out the message
    print(whole_message)

def length_validator(instance, attribute, value):
    '''Validator to ensure the attribute has exactly length 2.'''
    if len(value) != 2:
        raise ValueError(f"The '{attribute.name}' attribute must have a length of exactly 2.")

def copy_file(input_file_path, output_file_path):
    '''Use shutil to copy and the input file to the output location'''
    shutil.copy2(input_file_path, output_file_path)

def rename_file(rename_files_path, file):
    '''Use subprocess to run the executable file with the rename_files_path input'''
    subprocess.run([rename_files_path, file])

def convert_file(binary2asc_path, raw_file, ascii_file):
    '''Use os.system to run the binary2asc executable on the raw file and output the ascii_file'''
    cmd = f'{binary2asc_path} "{raw_file}" > "{ascii_file}"'
    os.system(cmd)

def create_tasks(working_directory,data_source,extension,output_data_dir,tasks,binary2asc_path):
    '''Create tasks for the ThreadPoolExecutor in the convert_binary_to_ascii function'''
    raw_files = working_directory.joinpath(data_source).joinpath(extension).rglob('*')
    for raw_file in raw_files:
        ascii_file = output_data_dir.joinpath(data_source, f'{raw_file.name}.asc')
        tasks.append((binary2asc_path, raw_file, ascii_file))
    return tasks

def read_sci_file(file:Path) -> pd.DataFrame:
    '''Tries to read a file from science and filters to select a few variables'''
    # Check if there are enough lines to read the file if there are not then return None,
    # any Nones are handled by the pd.concat function in join_ascii_files
    try:
        if os.path.getsize(file) > 0:
            # Read in the data
            df_raw = pd.read_csv(file, header=14, sep=' ', skiprows=[15,16])

            variables = df_raw.keys()
            # Define subsets of columns based on the presence of sci_oxy4_oxygen and sci_flbbcd_bb_units
            if 'sci_oxy4_oxygen' in variables and 'sci_flbbcd_bb_units' in variables:
                    present_variables = ['sci_m_present_time', 'sci_flbbcd_bb_units', 'sci_flbbcd_cdom_units', 'sci_flbbcd_chlor_units', 'sci_water_pressure', 'sci_water_temp', 'sci_water_cond', 'sci_oxy4_oxygen']

            elif 'sci_oxy4_oxygen' in variables and 'sci_flbbcd_bb_units' not in variables:
                    present_variables = ['sci_m_present_time', 'sci_water_pressure', 'sci_water_temp', 'sci_water_cond', 'sci_oxy4_oxygen']

            elif 'sci_oxy4_oxygen' not in variables and 'sci_flbbcd_bb_units' in variables:
                    present_variables = ['sci_m_present_time', 'sci_flbbcd_bb_units', 'sci_flbbcd_cdom_units', 'sci_flbbcd_chlor_units', 'sci_water_pressure', 'sci_water_temp', 'sci_water_cond']

            elif 'sci_oxy4_oxygen' not in variables and 'sci_flbbcd_bb_units' not in variables:
                    present_variables = ['sci_m_present_time', 'sci_water_pressure', 'sci_water_temp', 'sci_water_cond']

            df_filtered = df_raw[present_variables]
            # Parse the timestamp explicitly
            dates = df_filtered['sci_m_present_time'].copy()
            dates_datetime =  pd.to_datetime(dates, unit='s', errors='coerce')
            df_filtered = df_filtered.assign(sci_m_present_time=dates_datetime)
            df_filtered = df_filtered.set_index('sci_m_present_time')
        else:
            # Set the dataframe to 
            df_filtered = None
    except Exception as e:
        print(f'Unable to read and skipping: {file.stem} due to {e}')
        df_filtered = None
    
    return df_filtered

def read_flight_file(file:Path) -> pd.DataFrame:
    '''Tries to read flight data and filteres to select a few variables'''
    try:
        # Check if there are enough lines of data to read
        if os.path.getsize(file) > 0:
            df_raw = pd.read_csv(file, header=14, sep=' ', skiprows=[15,16])
            present_variables = ['m_present_time','m_lat','m_lon','m_pressure','m_water_depth']
            df_filtered = df_raw[present_variables]
            # Parse the timestamp explicitly
            dates = df_filtered['m_present_time'].copy()
            dates_datetime =  pd.to_datetime(dates, unit='s', errors='coerce')
            df_filtered = df_filtered.assign(m_present_time=dates_datetime)
            df_filtered = df_filtered.set_index(['m_present_time'])
        else:
            df_filtered = None
    except Exception as e:
        print(f'Unable to read and skipping: {file.stem} due to {e}')
        df_filtered = None
    
    return df_filtered

def join_ascii_files(files, file_reader, max_workers=None) -> pd.DataFrame:
    '''Uses ThreadPoolExecutor to read all files using a file reader function then concatenates all the files.'''
    if max_workers is None:
        max_workers = multiprocessing.cpu_count()

    # Use ThreadPoolExecutor to read files in parallel
    df_list = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_file = {executor.submit(file_reader, file): file for file in files}
        
        for future in as_completed(future_to_file):
            file = future_to_file[future]
            try:
                df_list.append(future.result())
            except Exception as e:
                print(f"Error reading {file}: {e}")
    df_list = [df for df in df_list if not df.empty]
    df_concat = pd.concat(df_list, axis=0)
    return df_concat

def process_sci_df(df:pd.DataFrame) -> pd.DataFrame:
    '''Process the data to filter and calculate salinity and density'''
    # Remove any data with erroneous dates (outside expected dates 2010-2030)
    upper_date_limit = str(datetime.datetime.today().date()+datetime.timedelta(days=365))
    df = df.reset_index()
    df = df[(df['sci_m_present_time'] > '2010-01-01') & (df['sci_m_present_time'] < upper_date_limit)]
    # Convert pressure from db to dbar
    df['sci_water_pressure'] = df['sci_water_pressure'] * 10
    # Calculate salinity and density
    df['sci_water_sal'] = gsw.SP_from_C(df['sci_water_cond']*10,df['sci_water_temp'],df['sci_water_pressure'])
    CT = gsw.CT_from_t(df['sci_water_sal'],df['sci_water_temp'],df['sci_water_pressure'])
    df['sci_water_dens'] = gsw.rho_t_exact(df['sci_water_sal'],CT,df['sci_water_pressure'])
    return df

def convert_sci_df_to_ds(df:pd.DataFrame,glider_id:dict,glider:str) -> xr.Dataset:
    '''Convert the given science dataframe to a xarray dataset'''
    bds = xr.Dataset() # put the platform info into the dataset on the top
    bds['platform'] = xr.DataArray(glider_id[glider])
    ds = xr.Dataset.from_dataframe(df)
    ds = bds.update(ds)
    return ds

def add_sci_attrs(ds:xr.Dataset,glider_id,glider,wmo_id) -> xr.Dataset:
    '''Add attributes to the science dataset'''
    variables = list(ds.data_vars)
    # Define variable attributes
    ds['platform'].attrs = {'ancillary_variables': ' ',
    'comment': ' ',
    'id': glider_id[glider],
    'instruments': 'instrument_ctd',
    'long_name': 'Slocum Glider '+ glider_id[glider],
    'type': 'platform',
    'wmo_id': wmo_id[glider],
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['sci_water_pressure'].attrs = {'accuracy': 0.01,
    'ancillary_variables': ' ',
    'axis': 'Z',
    'bytes': 4,
    'comment': 'Alias for sci_water_pressure, multiplied by 10 to convert from bar to dbar',
    'instrument': 'instrument_ctd',
    'long_name': 'CTD Pressure',
    'observation_type': 'measured',
    'platform': 'platform',
    'positive': 'down',
    'precision': 0.01,
    'reference_datum': 'sea-surface',
    'resolution': 0.01,
    'source_sensor': 'sci_water_pressure',
    'standard_name': 'sea_water_pressure',
    'units': 'bar',
    'valid_max': 2000.0,
    'valid_min': 0.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['sci_water_temp'].attrs = {'accuracy': 0.004,
    'ancillary_variables': ' ',
    'bytes': 4,
    'instrument': 'instrument_ctd',
    'long_name': 'Temperature',
    'observation_type': 'measured',
    'platform': 'platform',
    'precision': 0.001,
    'resolution': 0.001,
    'standard_name': 'sea_water_temperature',
    'units': 'Celsius',
    'valid_max': 40.0,
    'valid_min': -5.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['sci_water_cond'].attrs = {'accuracy': 0.001,
    'ancillary_variables': ' ',
    'bytes': 4,
    'instrument': 'instrument_ctd',
    'long_name': 'sci_water_cond',
    'observation_type': 'measured',
    'platform': 'platform',
    'precision': 1e-05,
    'resolution': 1e-05,
    'standard_name': 'sea_water_electrical_conductivity',
    'units': 'S m-1',
    'valid_max': 10.0,
    'valid_min': 0.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['sci_water_sal'].attrs = {'accuracy': ' ',
    'ancillary_variables': ' ',
    'instrument': 'instrument_ctd',
    'long_name': 'Salinity',
    'observation_type': 'calculated',
    'platform': 'platform',
    'precision': ' ',
    'resolution': ' ',
    'standard_name': 'sea_water_practical_salinity',
    'units': '1',
    'valid_max': 40.0,
    'valid_min': 0.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['sci_water_dens'].attrs = {'accuracy': ' ',
    'ancillary_variables': ' ',
    'instrument': 'instrument_ctd',
    'long_name': 'Density',
    'observation_type': 'calculated',
    'platform': 'platform',
    'precision': ' ',
    'resolution': ' ',
    'standard_name': 'sea_water_density',
    'units': 'kg m-3',
    'valid_max': 1040.0,
    'valid_min': 1015.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    if 'sci_flbbcd_bb_units' in variables:
        ds['sci_flbbcd_bb_units'].attrs = {'long_name':'science turbidity', 'standard_name':'backscatter', 'units':'nodim'}
        ds['sci_flbbcd_bb_units'].attrs = {'accuracy': ' ',
        'ancillary_variables': ' ',
        'instrument': 'instrument_flbbcd',
        'long_name': 'Turbidity',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': ' ',
        'resolution': ' ',
        'standard_name': 'sea_water_turbidity',
        'units': '1',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        ds['sci_flbbcd_cdom_units'].attrs = {'accuracy': ' ',
        'ancillary_variables': ' ',
        'instrument': 'instrument_flbbcd',
        'long_name': 'CDOM',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': ' ',
        'resolution': ' ',
        'standard_name': 'concentration_of_colored_dissolved_organic_matter_in_sea_water',
        'units': 'ppb',
        'valid_max': 50.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        ds['sci_flbbcd_chlor_units'].attrs = {'accuracy': ' ',
        'ancillary_variables': ' ',
        'instrument': 'instrument_flbbcd',
        'long_name': 'Chlorophyll_a',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': ' ',
        'resolution': ' ',
        'standard_name': 'mass_concentration_of_chlorophyll_a_in_sea_water',
        'units': '\u03BCg/L',
        'valid_max': 10.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}

    if 'sci_oxy4_oxygen' in variables:
        ds['sci_oxy4_oxygen'].attrs = {'accuracy': ' ',
        'ancillary_variables': ' ',
        'instrument': 'instrument_ctd_modular_do_sensor',
        'long_name': 'oxygen',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': ' ',
        'resolution': ' ',
        'standard_name': 'moles_of_oxygen_per_unit_mass_in_sea_water',
        'units': '\u03BCmol/kg',
        'valid_max': 500.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}

    return ds

def format_sci_ds(ds:xr.Dataset) -> xr.Dataset:
    '''Format the science dataset by sorting and renameing variables'''
    ds['index'] = np.sort(ds['sci_m_present_time'].values.astype('datetime64[s]'))
    ds = ds.drop_vars('sci_m_present_time')
    if 'sci_oxy4_oxygen' in ds.data_vars.keys():
        ds = ds.rename({'index': 'time','sci_water_pressure':'pressure','sci_water_temp':'temperature',
        'sci_water_cond':'conductivity','sci_water_sal':'salinity','sci_water_dens':'density','sci_flbbcd_bb_units':'turbidity',
        'sci_flbbcd_cdom_units':'cdom','sci_flbbcd_chlor_units':'chlorophyll','sci_oxy4_oxygen':'oxygen'})
    else:

        ds = ds.rename({'index': 'time','sci_water_pressure':'pressure','sci_water_temp':'temperature',
        'sci_water_cond':'conductivity','sci_water_sal':'salinity','sci_water_dens':'density'})
    return ds

def process_flight_df(df:pd.DataFrame) -> pd.DataFrame:
    '''Process flight dataframe by filtering and calculating latitude and longitude and renaming variables'''
    # Exclude data before 2020 and after 2030
    df = df.reset_index()
    df = df[(df['m_present_time'] > '2010-01-01') & (df['m_present_time'] < '2030-01-01')]
    # Convert pressure from db to dbar
    df['m_pressure'] = df['m_pressure'] * 10
    # Convert latitude and longitude to decimal degrees
    # .round(0) will up round the decimal number > 0.5 to 1, which is not working
    df['m_lat'] = df['m_lat'] / 100.0
    df['m_lat'] = np.sign(df['m_lat'])*(np.sign(df['m_lat'])*df['m_lat']-(np.sign(df['m_lat'])*df['m_lat'])%1 + (np.sign(df['m_lat'])*df['m_lat'])%1/0.6)

    df['m_lon'] = df['m_lon'] / 100.0
    df['m_lon'] = np.sign(df['m_lon'])*(np.sign(df['m_lon'])*df['m_lon']-(np.sign(df['m_lon'])*df['m_lon'])%1 + (np.sign(df['m_lon'])*df['m_lon'])%1/0.6)

    # Rename columns for clarity
    df.rename(columns={'m_lat': 'm_latitude', 'm_lon': 'm_longitude'}, inplace=True)

    return df

def convert_fli_df_to_ds(df:pd.DataFrame) -> xr.Dataset:
    '''Convert the flight dataframe to dataset'''
    ds = xr.Dataset.from_dataframe(df)
    return ds

def add_flight_attrs(ds:xr.Dataset) -> xr.Dataset:
    '''Add attributes to the flight dataset'''
    ds['m_pressure'].attrs = {'accuracy': 0.01,
    'ancillary_variables': ' ',
    'axis': 'Z',
    'bytes': 4,
    'comment': 'Alias for m_pressure, multiplied by 10 to convert from bar to dbar',
    'long_name': 'GPS Pressure',
    'observation_type': 'measured',
    'platform': 'platform',
    'positive': 'down',
    'precision': 0.01,
    'reference_datum': 'sea-surface',
    'resolution': 0.01,
    'source_sensor': 'sci_water_pressure',
    'standard_name': 'sea_water_pressure',
    'units': 'bar',
    'valid_max': 2000.0,
    'valid_min': 0.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['m_water_depth'].attrs = {'accuracy': 0.01,
    'ancillary_variables': ' ',
    'axis': 'Z',
    'bytes': 4,
    'comment': 'Alias for m_depth',
    'long_name': 'GPS Depth',
    'observation_type': 'calculated',
    'platform': 'platform',
    'positive': 'down',
    'precision': 0.01,
    'reference_datum': 'sea-surface',
    'resolution': 0.01,
    'source_sensor': 'm_depth',
    'standard_name': 'sea_water_depth',
    'units': 'meters',
    'valid_max': 2000.0,
    'valid_min': 0.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['m_latitude'].attrs = {'ancillary_variables': ' ',
    'axis': 'Y',
    'bytes': 8,
    'comment': 'm_gps_lat converted to decimal degrees and interpolated',
    'coordinate_reference_frame': 'urn:ogc:crs:EPSG::4326',
    'long_name': 'Latitude',
    'observation_type': 'calculated',
    'platform': 'platform',
    'precision': 5,
    'reference': 'WGS84',
    'source_sensor': 'm_gps_lat',
    'standard_name': 'latitude',
    'units': 'degree_north',
    'valid_max': 90.0,
    'valid_min': -90.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
    ds['m_longitude'].attrs = {'ancillary_variables': ' ',
    'axis': 'X',
    'bytes': 8,
    'comment': 'm_gps_lon converted to decimal degrees and interpolated',
    'coordinate_reference_frame': 'urn:ogc:crs:EPSG::4326',
    'long_name': 'Longitude',
    'observation_type': 'calculated',
    'platform': 'platform',
    'precision': 5,
    'reference': 'WGS84',
    'source_sensor': 'm_gps_lon',
    'standard_name': 'longitude',
    'units': 'degree_east',
    'valid_max': 180.0,
    'valid_min': -180.0,
    'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}

    return ds

def format_flight_ds(ds:xr.Dataset) -> xr.Dataset:
    '''Format the flight dataset by sorting and renaming variables'''
    ds['index'] = np.sort(ds['m_present_time'].values.astype('datetime64[s]'))
    ds = ds.drop_vars('m_present_time')
    ds = ds.rename({'index': 'm_time','m_pressure':'m_pressure','m_water_depth':'depth','m_latitude':'latitude','m_longitude':'longitude'})

    return ds

def process_sci_data(science_data_dir,glider_id,glider,wmo_id) -> xr.Dataset:
    '''Perform all processing of science data from ascii to pandas dataframe to xarray dataset'''
    print_time('Processing Science Data')
    # Process Science Data
    sci_files = list(science_data_dir.rglob("*.asc"))
    sci_files.sort()
    df_sci = join_ascii_files(sci_files,read_sci_file)
    df_sci = process_sci_df(df_sci)
    ds_sci = convert_sci_df_to_ds(df_sci,glider_id,glider)
    ds_sci = add_sci_attrs(ds_sci,glider_id,glider,wmo_id)
    ds_sci = format_sci_ds(ds_sci)
    print_time('Finished Processing Science Data')
    return ds_sci

def process_flight_data(flight_data_dir) -> xr.Dataset:
    '''Perform all processing of flight data from ascii to pandas dataframe to xarray dataset'''
    print_time('Processing Flight Data')
    # Process Flight Data
    fli_files = list(flight_data_dir.rglob("*.asc"))
    fli_files.sort()
    df_fli = join_ascii_files(fli_files,read_flight_file)
    df_fli = process_flight_df(df_fli)
    ds_fli = convert_fli_df_to_ds(df_fli)
    ds_fli = add_flight_attrs(ds_fli)
    ds_fli = format_flight_ds(ds_fli)
    print_time('Finised Processing Flight Data')
    return ds_fli

def add_gridded_data(ds_mission:xr.Dataset) -> xr.Dataset:
    '''Create gridder object and create the gridded dataset'''
    print_time('Adding Gridded Data')
    gridder = Gridder(ds_mission=ds_mission)
    gridder.create_gridded_dataset()
    ds_mission.update(gridder.ds_gridded)
    print_time('Finished Adding Gridded Data')
    return ds_mission

def get_polygon_coords(ds_mission:xr.Dataset) -> str:
    '''Get the polygon coords for the dataset global attributes'''
    lat_max = np.nanmax(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)
    lat_min = np.nanmin(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)
    lon_max = np.nanmax(ds_mission.longitude.values)
    lon_min = np.nanmin(ds_mission.longitude.values)
    polygon_1 = str(lat_max)+' '+str(ds_mission.longitude[np.where(ds_mission.latitude==lat_max)[0][0]].values) # northmost
    polygon_2 = str(ds_mission.latitude[np.where(ds_mission.longitude==lon_max)[0][0]].values)+' '+str(lon_max) # eastmost
    polygon_3 = str(lat_min)+' '+str(ds_mission.longitude[np.where(ds_mission.latitude==lat_min)[0][0]].values) # southmost
    polygon_4 = str(ds_mission.latitude[np.where(ds_mission.longitude==lon_min)[0][0]].values)+' '+str(lon_min) # westmost
    polygon_5 = polygon_1
    return 'POLYGON (('+polygon_1+' '+polygon_2+' '+polygon_3+' '+polygon_4+' '+polygon_5+'))'

def add_global_attrs(ds_mission:xr.Dataset,mission_title:str,wmo_id:dict,glider:str) -> xr.Dataset:
    '''Add attributes to the mission dataset'''
    ds_mission.attrs = {'Conventions': 'CF-1.6, COARDS, ACDD-1.3',
    'acknowledgment': ' ',
    'cdm_data_type': 'Profile',
    'comment': 'time is the ctd_time from sci_m_present_time, m_time is the gps_time from m_present_time, g_time and g_pres are the grided time and pressure',
    'contributor_name': 'Steven F. DiMarco',
    'contributor_role': ' ',
    'creator_email': 'sakib@tamu.edu, gexiao@tamu.edu',
    'creator_institution': 'Texas A&M University, Geochemical and Environmental Research Group',
    'creator_name': 'Sakib Mahmud, Xiao Ge',
    'creator_type': 'persons',
    'creator_url': 'https://gerg.tamu.edu/',
    'date_created': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S'),
    'date_issued': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S'),
    'date_metadata_modified': '2023-09-15',
    'date_modified': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S'),
    'deployment': ' ',
    'featureType': 'profile',
    'geospatial_bounds_crs': 'EPSG:4326',
    'geospatial_bounds_vertical_crs': 'EPSG:5831',
    'geospatial_lat_resolution': "{:.4e}".format(abs(np.nanmean(np.diff(ds_mission.latitude))))+ ' degree',
    'geospatial_lat_units': 'degree_north',
    'geospatial_lon_resolution': "{:.4e}".format(abs(np.nanmean(np.diff(ds_mission.longitude))))+ ' degree',
    'geospatial_lon_units': 'degree_east',
    'geospatial_vertical_positive': 'down',
    'geospatial_vertical_resolution': ' ',
    'geospatial_vertical_units': 'EPSG:5831',
    'infoUrl': 'https://gerg.tamu.edu/',
    'institution': 'Texas A&M University, Geochemical and Environmental Research Group',
    'instrument': 'In Situ/Laboratory Instruments > Profilers/Sounders > CTD',
    'instrument_vocabulary': 'NASA/GCMD Instrument Keywords Version 8.5',
    'ioos_regional_association': 'GCOOS-RA',
    'keywords': 'Oceans > Ocean Pressure > Water Pressure, Oceans > Ocean Temperature > Water Temperature, Oceans > Salinity/Density > Conductivity, Oceans > Salinity/Density > Density, Oceans > Salinity/Density > Salinity',
    'keywords_vocabulary': 'NASA/GCMD Earth Sciences Keywords Version 8.5',
    'license': 'This data may be redistributed and used without restriction.  Data provided as is with no expressed or implied assurance of quality assurance or quality control',
    'metadata_link': ' ',
    'naming_authority': 'org.gcoos.gandalf',
    'ncei_template_version': 'NCEI_NetCDF_Trajectory_Template_v2.0',
    'platform': 'In Situ Ocean-based Platforms > AUVS > Autonomous Underwater Vehicles',
    'platform_type': 'Slocum Glider',
    'platform_vocabulary': 'NASA/GCMD Platforms Keywords Version 8.5',
    'processing_level': 'Level 0',
    'product_version': '0.0',
    'program': ' ',
    'project': ' ',
    'publisher_email': 'sdimarco@tamu.edu',
    'publisher_institution': 'Texas A&M University, Geochemical and Environmental Research Group',
    'publisher_name': 'Steven F. DiMarco',
    'publisher_url': 'https://gerg.tamu.edu/',
    'references': ' ',
    'sea_name': 'Gulf of Mexico',
    'standard_name_vocabulary': 'CF Standard Name Table v27',
    'summary': 'Merged dataset for GERG future usage.',
    'time_coverage_resolution': ' ',
    'wmo_id': wmo_id[glider],
    'uuid': str(uuid.uuid4()),
    'history': 'dbd and ebd files transferred from dbd2asc on 2023-09-15, merged into single netCDF file on '+pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S'),
    'title': mission_title,
    'source': 'Observational Slocum glider data from source ebd and dbd files',
    'geospatial_lat_min': str(np.nanmin(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)),
    'geospatial_lat_max': str(np.nanmax(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)),
    'geospatial_lon_min': str(np.nanmin(ds_mission.longitude.values)),
    'geospatial_lon_max': str(np.nanmax(ds_mission.longitude.values)),
    'geospatial_bounds': get_polygon_coords(ds_mission),
    'geospatial_vertical_min': str(np.nanmin(ds_mission.depth[np.where(ds_mission.depth>0)].values)),
    'geospatial_vertical_max': str(np.nanmax(ds_mission.depth.values)),
    'time_coverage_start': str(ds_mission.time[-1].values)[:19],
    'time_coverage_end': str(ds_mission.m_time[-1].values)[:19],
    'time_coverage_duration': 'PT'+str((ds_mission.m_time[-1].values - ds_mission.time[-1].values) / np.timedelta64(1, 's'))+'S'}

    return ds_mission

def save_ds(ds_mission:xr.Dataset,output_nc_path):
    '''Save xarray.Dataset to NetCDF'''
    print_time('Saving Dataset to NetCDF')
    ds_mission.to_netcdf(output_nc_path)
    print_time('Done Saving Dataset to NetCDF')

def convert_ascii_to_dataset(working_directory:Path,glider:str,mission_title:str):
    '''Convert ascii data files into a single NetCDF file'''
    print_time('Converting ascii to netcdf')
    working_directory = working_directory.joinpath('processed')

    science_data_dir:Path = working_directory.joinpath('Science')
    flight_data_dir:Path = working_directory.joinpath('Flight')

    # output_nc_path = working_directory.joinpath('processed','nc',nc_filename)

    glider_id = {'199':'Dora','307':'Reveille','308':'Howdy','540':'Stommel','541':'Sverdrup'}
    wmo_id = {'199':'unknown','307':'4801938','308':'4801915','540':'4801916','541':'4801924'}
    
    # Process Science Data
    ds_sci:xr.Dataset = process_sci_data(science_data_dir,glider_id,glider,wmo_id)

    # Make a copy of the science dataset
    ds_mission:xr.Dataset = ds_sci.copy()

    # Process Flight Data
    ds_fli:xr.Dataset = process_flight_data(flight_data_dir)

    # Add flight data to mission dataset
    ds_mission.update(ds_fli)

    # Add gridded data to mission dataset
    ds_mission = add_gridded_data(ds_mission)

    # Add attributes to the mission dataset
    ds_mission = add_global_attrs(ds_mission,mission_title=mission_title,wmo_id=wmo_id,glider=glider)

    print_time('Finished converting ascii to dataset')
    return ds_mission




'''
Module containing the Gridder class
'''
from attrs import define,field
import numpy as np
import pandas as pd
import xarray as xr
import gsw

@define
class Gridder:
    '''
    Class to create and calculate the gridded dataset.
    This class handles some functions that are coupled to make the code easier to read and maintain
    '''
    ds_mission:xr.Dataset
    interval_h:int|float = field(default=1)
    interval_p:int|float = field(default=0.1)

    ds:xr.Dataset = field(init=False)
    ds_gridded:xr.Dataset = field(init=False)
    variable_names:list = field(init=False)
    time:np.ndarray = field(init=False)
    pres:np.ndarray = field(init=False)
    lat:np.ndarray = field(init=False)
    lon:np.ndarray = field(init=False)
    xx:np.ndarray = field(init=False)
    yy:np.ndarray = field(init=False)
    int_time:np.ndarray = field(init=False)
    int_pres:np.ndarray = field(init=False)
    data_arrays:dict = field(init=False)
    grid_pres:np.ndarray = field(init=False)
    grid_time:np.ndarray = field(init=False)

    def __attrs_post_init__(self):
        '''
        Initalize the class and get the dimension values of the dataset
        '''
        self.ds = self.ds_mission.copy()
        # Get indexes of where there are non-nan pressure values
        tloc_idx = np.where(~np.isnan(self.ds['pressure']))[0]
        # Select the times were thre are non-nan pressure values
        self.ds = self.ds.isel(time=tloc_idx)
        # Get all of the variables present in the dataset
        self.variable_names = list(self.ds.data_vars.keys())
        # Get all of the time values in the dataset
        self.time = self.ds.time.values
        # Get all of the pressure values in the dataset
        self.pres = self.ds.pressure.values
        # Get all of the lat and lon values in the dataset
        self.lon = np.nanmean(self.ds_mission.longitude)
        self.lat = np.nanmean(self.ds_mission.latitude[np.where(self.ds_mission.latitude.values<29.5)].values)

        self.initalize_grid()
    
    def initalize_grid(self):
        '''
        Initalize the grid that is used to calcuate the gridded dataset
        '''
        start_hour = int(pd.to_datetime(self.time[0]).hour / self.interval_h) * self.interval_h
        end_hour = int(pd.to_datetime(self.time[-1]).hour / self.interval_h) * self.interval_h
        start_time = pd.to_datetime(self.time[0]).replace(hour=start_hour, minute=0, second=0)
        end_time = pd.to_datetime(self.time[-1]).replace(hour=end_hour, minute=0, second=0)

        self.int_time = np.arange(start_time, end_time+np.timedelta64(self.interval_h, 'h'), np.timedelta64(self.interval_h, 'h')).astype('datetime64[s]')

        # create the pressure grids for intepolation
        start_pres = 0
        end_pres = np.nanmax(self.pres)
        self.int_pres = np.arange(start_pres, end_pres, self.interval_p)

        self.grid_pres,self.grid_time = np.meshgrid(self.int_pres,self.int_time[1:]) # get the time between two time point
        self.xx,self.yy = np.shape(self.grid_pres)

        # List of variable names
        var_names = ['int_temp', 'int_salt', 'int_cond', 'int_dens', 'int_turb', 'int_cdom', 'int_chlo', 'int_oxy4']

        # Dictionary to store the arrays
        self.data_arrays = {}

        # Initialize each array with NaN values and store it in the dictionary
        for var in var_names:
            self.data_arrays[var] = np.empty((self.xx, self.yy))
            self.data_arrays[var].fill(np.nan)

    def add_attrs(self):
        '''
        Add dataset attributes to the gridded variables
        '''
        self.ds_gridded['g_temp'].attrs = {'long_name': 'Gridded Temperature',
        'observation_type': 'calculated',
        'source': 'temperature from sci_water_temp',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_temperature',
        'units': 'Celsius',
        'valid_max': 40.0,
        'valid_min': -5.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_salt'].attrs = {'long_name': 'Gridded Salinity',
        'observation_type': 'calculated',
        'source': 'salinity from sci_water_sal',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_practical_salinity',
        'units': '1',
        'valid_max': 40.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_cond'].attrs = {'long_name': 'Gridded Conductivity',
        'observation_type': 'calculated',
        'source': 'conductivity from sci_water_cond',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_electrical_conductivity',
        'units': 'S m-1',
        'valid_max': 10.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_dens'].attrs = {'long_name': 'Gridded Density',
        'observation_type': 'calculated',
        'source': 'density from sci_water_dens',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_density',
        'units': 'kg m-3',
        'valid_max': 1040.0,
        'valid_min': 1015.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        if 'bb' in self.variable_names:
            self.ds_gridded['g_turb'].attrs = {'long_name': 'Gridded Turbidity',
            'observation_type': 'calculated',
            'source': 'turbidity from sci_flbbcd_bb_units',
            'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
            'standard_name': 'sea_water_turbidity',
            'units': '1',
            'valid_max': 1.0,
            'valid_min': 0.0,
            'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        if 'cdom' in self.variable_names:
            self.ds_gridded['g_cdom'].attrs = {'long_name': 'Gridded CDOM',
            'observation_type': 'calculated',
            'source': 'cdom from sci_flbbcd_cdom_units',
            'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
            'standard_name': 'concentration_of_colored_dissolved_organic_matter_in_sea_water',
            'units': 'ppb',
            'valid_max': 50.0,
            'valid_min': 0.0,
            'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        if 'chlor' in self.variable_names:
            self.ds_gridded['g_chlo'].attrs = {'long_name': 'Gridded Chlorophyll_a',
            'observation_type': 'calculated',
            'source': 'chlorophyll from sci_flbbcd_chlor_units',
            'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
            'standard_name': 'mass_concentration_of_chlorophyll_a_in_sea_water',
            'units': '\u03BCg/L',
            'valid_max': 10.0,
            'valid_min': 0.0,
            'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        if 'oxygen' in self.variable_names:
            self.ds_gridded['g_oxy4'].attrs = {'long_name': 'Gridded Oxygen',
            'observation_type': 'calculated',
            'source': 'oxygen from sci_oxy4_oxygen',
            'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
            'standard_name': 'moles_of_oxygen_per_unit_mass_in_sea_water',
            'units': '\u03BCmol/kg',
            'valid_max': 500.0,
            'valid_min': 0.0,
            'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_hc'].attrs = {'long_name': 'Gridded Heat Content',
        'observation_type': 'calculated',
        'source': 'g_temp, g_dens, cp=gsw.cp_t_exact, dz='+str(self.interval_p)+'dbar',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_heat_content_for_all_grids',
        'units': 'kJ/cm^2',
        'valid_max': 10.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_phc'].attrs = {'long_name': 'Gridded Potential Heat Content',
        'observation_type': 'calculated',
        'source': 'g_temp, g_dens, cp=gsw.cp_t_exact, dz='+str(self.interval_p)+'dbar',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_heat_content_for_grids_above_26°C',
        'units': 'kJ/cm^2',
        'valid_max': 10.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_sp'].attrs = {'long_name': 'Gridded Spiciness',
        'observation_type': 'calculated',
        'source': 'g_temp, g_salt, g_pres, lon, lat, via gsw.spiciness0',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'spiciness_from_absolute_salinity_and_conservative_temperature_at_0dbar',
        'units': 'kg/m^3',
        'valid_max': 10.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.ds_gridded['g_depth'].attrs = {'long_name': 'Gridded Depth',
        'observation_type': 'calculated',
        'source': 'g_pres, lat, via gsw.z_from_p',
        'resolution': str(self.interval_h)+'hour and '+str(self.interval_p)+'dbar',
        'standard_name': 'sea_water_depth',
        'units': 'm',
        'valid_max': 1000.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}

    def create_gridded_dataset(self):
        '''
        Calculate gridded data and create gridded dataset
        '''
        for ttt in range(self.xx):
            tds:xr.Dataset = self.ds.sel(time=slice(str(self.int_time[ttt]),str(self.int_time[ttt+1]))).copy()
            if len(tds.time) > 0:
                tds = tds.sortby('pressure')
                tds = tds.assign_coords(time=('time',tds.time.values.astype('datetime64[s]')))
                tds['time'] = tds['pressure'].values
                
                # Remove duplicates and slightly modify if necessary by adding a tiny value
                unique_pressures, indices, counts = np.unique(tds['pressure'], return_index=True, return_counts=True)
                duplicates = unique_pressures[counts > 1]
                for pressure in duplicates:
                    rrr = np.where(tds['pressure'] == pressure)[0]
                    for rr in rrr:
                        modified_pressure = pressure + 0.000000000001*rr
                        tds['pressure'][rr] = modified_pressure
                tds['time'] = tds['pressure'].values # save new non-duplicates pressure
                self.data_arrays['int_temp'][ttt,:] = tds.temperature.interp(time=self.int_pres)
                self.data_arrays['int_salt'][ttt,:] = tds.salinity.interp(time=self.int_pres)
                self.data_arrays['int_cond'][ttt,:] = tds.conductivity.interp(time=self.int_pres)
                self.data_arrays['int_dens'][ttt,:] = tds.density.interp(time=self.int_pres)
                if 'oxygen' in self.variable_names:
                    self.data_arrays['int_oxy4'][ttt,:] = tds.oxygen.interp(time=self.int_pres)

        # give a dz instead of calculating the inter depth
        sa = gsw.SA_from_SP(self.data_arrays['int_salt'], self.grid_pres, self.lon, self.lat)
        pt = gsw.pt0_from_t(sa, self.data_arrays['int_temp'], self.grid_pres)
        ct = gsw.CT_from_pt(sa, pt)
        cp = gsw.cp_t_exact(sa, self.data_arrays['int_temp'], self.grid_pres) * 0.001 # from J/(kg*K) to kJ/(kg*°C) or use 3.85 as a constant?
        dep = gsw.z_from_p(self.grid_pres, self.lat, geo_strf_dyn_height=0, sea_surface_geopotential=0)
        spc = gsw.spiciness0(sa, ct)

        dz = self.interval_p
        hc = dz*cp*self.data_arrays['int_temp']*self.data_arrays['int_dens'] # deltaZ * Cp * temperature * density in the unit as $[kJ/m^2]$ * 10**-4 to $[kJ/cm^2]$
        phc = dz*cp*(self.data_arrays['int_temp']-26)*self.data_arrays['int_dens'] # deltaZ * Cp * temperature * density in the unit as $[kJ/m^2]$ * 10**-4 to $[kJ/cm^2]$
        phc[phc<0] = np.nan
        self.ds_gridded = xr.Dataset()
        self.ds_gridded['g_temp'] = xr.DataArray(self.data_arrays['int_temp'],[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_salt'] = xr.DataArray(self.data_arrays['int_salt'],[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_cond'] = xr.DataArray(self.data_arrays['int_cond'],[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_dens'] = xr.DataArray(self.data_arrays['int_dens'],[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        if 'oxygen' in self.variable_names:
            self.ds_gridded['g_oxy4'] = xr.DataArray(self.data_arrays['int_oxy4'],[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_hc'] = xr.DataArray(hc*10**-4,[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_phc'] = xr.DataArray(phc*10**-4,[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_sp'] = xr.DataArray(spc,[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])
        self.ds_gridded['g_depth'] = xr.DataArray(dep,[('g_time',self.int_time[1:]),('g_pres',self.int_pres)])

        self.add_attrs()
